/*
 * This file is part of the AdvanceMAME project.
 *
 * Copyright (C) 1999-2002 Andrea Mazzoleni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "video.h"
#include "draw.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

extern void report(void);

struct result {
	char driver[VIDEO_NAME_MAX];
	video_crtc crtc;
	double vmeasured;
	double vexpected;
	double pexpected;
	int set;
};

struct result result_map[128];
unsigned result_mac;

video_error bug_test(const char* modeline) {
	video_crtc crtc;
	video_mode mode;

	unsigned i;

	video_log("report: test modeline \"%s\"\n",modeline);

	if (video_crtc_parse(&crtc, modeline, modeline + strlen(modeline)) != 0) {
		video_log("report: video_crtc_parse() failed\n");
		return -1;
	}

	i = result_mac++;

	strcpy(result_map[i].driver,"unknow");
	result_map[i].crtc = crtc;
	result_map[i].vexpected = 0;
	result_map[i].vmeasured = 0;
	result_map[i].pexpected = 0;
	result_map[i].set = 0;

	if (video_mode_generate(&mode, &crtc, 8, VIDEO_FLAGS_TYPE_GRAPHICS | VIDEO_FLAGS_INDEX_RGB) != 0) {
		video_log("report: video_mode_generate() failed\n");
		return -1;
	}

	strcpy(result_map[i].driver,video_mode_driver(&mode)->name);
	strcat(result_map[i].driver,"/");
	strcat(result_map[i].driver,video_mode_driver(&mode)->driver_name());

	if (video_mode_set(&mode)!=0) {
		video_log("report: video_mode_set() failed\n");
	}

	usleep(200000);

	result_map[i].vexpected = video_vclock();
	result_map[i].vmeasured = video_measured_vclock();
	result_map[i].pexpected = crtc_pclock_get(&crtc);
	result_map[i].set = 1;

	video_mode_done();

	usleep(200000);

	return 0;
}

const char* MODELINE[] = {
"pc_31.5_320x240 12.5876 320 328 376 400 240 245 246 262 -hsync -vsync doublescan", /* H 31469 [Hz], V 60.0553 [Hz] */
"pc_31.5_640x480 25.1752 640 656 752 800 480 490 492 525 -hsync -vsync", /* H 31469 [Hz], V 59.941 [Hz] */
"pc_31.5_1024x768 40.2803 1024 1048 1200 1280 768 784 787 840 -hsync -vsync interlace", /* H 31469 [Hz], V 74.9262 [Hz] */
"ntsc_320x242 6.04186 320 328 352 384 242 245 248 262 -hsync -vsync", /* H 15734 [Hz], V 60.0534 [Hz] */
"ntsc_640x242 12.0837 640 656 704 768 242 245 248 262 -hsync -vsync", /* H 15734 [Hz], V 60.0534 [Hz] */
"ntsc_640x484 12.0837 640 656 704 768 484 490 496 525 -hsync -vsync interlace", /* H 15734 [Hz], V 59.939 [Hz] */
"pal_384x288 7.375 384 392 432 472 288 291 294 312 -hsync -vsync", /* H 15625 [Hz], V 50.0801 [Hz] */
"pal_768x288 14.75 768 784 864 944 288 291 294 312 -hsync -vsync", /* H 15625 [Hz], V 50.0801 [Hz] */
"pal_768x576 14.75 768 784 864 944 576 582 588 625 -hsync -vsync interlace", /* H 15625 [Hz], V 50 [Hz] */
"standard_320x240 6.79104 320 352 384 432 240 243 246 262 -hsync -vsync", /* H 15720 [Hz], V 60 [Hz] */
"standard_640x240 13.5821 640 704 768 864 240 243 246 262 -hsync -vsync", /* H 15720 [Hz], V 60 [Hz] */
"standard_640x480 13.5821 640 704 768 864 480 486 493 524 -hsync -vsync interlace", /* H 15720 [Hz], V 60 [Hz] */
0
};

void report(void)
{
	const char** modeline;
	int key;
	unsigned i;
	double pclock_min = 0;

	sound_error();

	printf("PLEASE READ CAREFULLY and follow the steps below:\n");
	printf("\n");
	printf("1) POWER OFF your monitor.\n");
	printf("2) Press ENTER.\n");
	printf("3) Wait 20 seconds and you should hear a sound.\n");
	printf("4) POWER ON your monitor.\n");
	printf("\n");
	printf("If after 20 seconds you don't hear any sound RESET the PC.\n");
	printf("DON'T POWER ON the monitor until you hear the termination sound!\n");
	printf("\n");
	printf("Now POWER OFF the monitor and press ENTER to continue or ESC to abort.\n");

	do {
		key = getch();
	} while (key != 13 && key != 27);
	if (key != 13) {
		printf("Abort.\n");
		exit(EXIT_FAILURE);
	}

	video_log("\nPLEASE SEND this file to advancemame-report@lists.sourceforge.net\n\n");

	video_log("report: %s %s\n",__DATE__,__TIME__);

	video_log("report: test svgaline\n");

	video_register(&svgaline_driver);
	video_init();
	modeline = (const char**)&MODELINE;
	while (*modeline) {
		bug_test(*modeline);
		++modeline;
	}
	video_mode_reset();
	video_done();

	video_log("report: test vbeline\n");

	video_register(&vbeline_driver);
	video_init();
	modeline = (const char**)&MODELINE;
	while (*modeline) {
		bug_test(*modeline);
		++modeline;
	}
	video_mode_reset();
	video_done();

	sound_error();
	sound_warn();
	sound_error();
	sound_warn();

	printf("Result summary:\n\n");
	printf("%-33s | %-18s | %-18s\n","modeline", result_map[0].driver, result_map[result_mac/2].driver );
	video_log("summary: %-33s | %-18s | %-18s\n","modeline", result_map[0].driver, result_map[result_mac/2].driver );
	for(i=0;i<result_mac / 2;++i) {
		char buffer0[256];
		char buffer1[256];
		char buffer2[256];
		const char* flags;
		unsigned j = i + result_mac / 2;

		if (result_map[i].crtc.interlaced)
			flags = "interlace";
		else if (result_map[i].crtc.doublescan)
			flags = "doublescan";
		else
			flags = "singlescan";

		sprintf(buffer0,"%-16s %5.1f %-10s",  result_map[i].crtc.name, result_map[i].vexpected, flags);

		if (result_map[i].set) {
			const char* msg;
			double factor;
			if (result_map[i].vmeasured) {
				factor = result_map[i].vexpected / result_map[i].vmeasured;
				if (factor > 1.0)
					factor = 1.0 / factor;
				if (factor < 0.95) {
					msg = "ERROR";
					if (pclock_min == 0) {
						pclock_min = result_map[i].pexpected / factor;
					}
				} else
					msg = "OK";
			} else
				msg = "ERROR";
			sprintf(buffer1,"%5.1f %s", result_map[i].vmeasured, msg);
		} else {
			sprintf(buffer1,"FAILED");
		}

		if (result_map[j].set) {
			const char* msg;
			double factor;
			if (result_map[j].vmeasured) {
				factor = result_map[j].vexpected / result_map[j].vmeasured;
				if (factor > 1.0)
					factor = 1.0 / factor;
				if (factor < 0.95)
					msg = "ERROR";
				else
					msg = "OK";
			} else
				msg = "ERROR";
			sprintf(buffer2,"%5.1f %s", result_map[j].vmeasured, msg);
		} else {
			sprintf(buffer2,"FAILED");
		}

		printf("%-33s | %-18s | %-18s\n",buffer0,buffer1,buffer2);
		video_log("summary: %-33s | %-18s | %-18s\n",buffer0,buffer1,buffer2);

	}

	if (pclock_min != 0) {
		printf("\n(svgaline lower pclock is %g MHz)\n",pclock_min / 1E6);
		video_log("summary:(svgaline lower pclock is %g MHz)\n",pclock_min / 1E6);
	}

	printf("\nPLEASE SEND the file 'report.log' to advancemame-report@lists.sourceforge.net\n");
	video_log("\nPLEASE SEND this file to advancemame-report@lists.sourceforge.net\n");
}
