/*
 * This file is part of the AdvanceMAME project.
 *
 * Copyright (C) 1999-2002 Andrea Mazzoleni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

unsigned read_be_u32(unsigned char* v) {
	return v[0] << 24 | v[1] << 16 | v[2] << 8 | v[3];
}

unsigned read_be_u16(unsigned char* v) {
	return v[0] << 8 | v[1];
}

int fread_be_u32(FILE* f, unsigned* r) {
	unsigned char v[4];
	if (fread(v,4,1,f) != 1) {
		return -1;
	}
	*r = (unsigned)v[0] << 24 | (unsigned)v[1] << 16 | (unsigned)v[2] << 8 | (unsigned)v[3];

	return 0;
}

int fread_tag(FILE* f, char* v) {
	if (fread(v,4,1,f) != 1) {
		return -1;
	}
	return 0;
}

int fskip(FILE*f, unsigned n) {
	while (n > 0) {
		char c;
		if (fread(&c,1,1,f) != 1) {
			return -1;
		}
		--n;
	}

	return 0;
}

#define PNG_CN_IHDR 0x49484452
#define MNG_CN_DHDR 0x44484452
#define MNG_CN_MHDR 0x4D484452
#define MNG_CN_DEFI 0x44454649
#define MNG_CN_PPLT 0x50504c54
#define MNG_CN_MOVE 0x4d4f5645
#define MNG_CN_FRAM 0x4652414d

void print(unsigned char* tag, unsigned char* data, unsigned size) {
	unsigned type = read_be_u32(tag);
	unsigned i;

	printf("%c%c%c%c%6d",tag[0],tag[1],tag[2],tag[3],size);

	switch (type) {
		case MNG_CN_MHDR :
			printf(" width:%d height:%d tick:%d simplicity:%08xh", read_be_u32(data+0), read_be_u32(data+4), read_be_u32(data+8), read_be_u32(data+24) );
		break;
		case MNG_CN_DHDR :
			printf(" id:%d", read_be_u16(data+0));
			switch (data[2]) {
				case 0 : printf(" img:unspecified"); break;
				case 1 : printf(" img:png"); break;
				case 2 : printf(" img:jng"); break;
				default: printf(" img:?"); break;
			}
			switch (data[3]) {
				case 0 : printf(" delta:entire_replacement"); break;
				case 1 : printf(" delta:block_addition"); break;
				case 2 : printf(" delta:block_alpha_addition"); break;
				case 3 : printf(" delta:block_color_addition"); break;
				case 4 : printf(" delta:block_replacement"); break;
				case 5 : printf(" delta:block_alpha_replacement"); break;
				case 6 : printf(" delta:block_color_replacement"); break;
				case 7 : printf(" delta:no_change"); break;
				default: printf(" delta:?"); break;
			}
			if (size >= 12) {
				printf(" width:%d height:%d", read_be_u32(data + 4), read_be_u32(data + 8) );
			}
			if (size >= 20) {
				printf(" x:%d y:%d", read_be_u32(data + 12), read_be_u32(data + 16) );
			}
		break;
		case MNG_CN_FRAM :
			if (size >= 1) {
				printf(" mode:%d",(unsigned)data[0]);
			}
			if (size > 1) {
				i = 1;
				while (i < size && data[i]!=0)
					++i;
				printf(" len:%d",i-1);

				if (size >= i+2) {
					printf(" delay_mode:%d",(unsigned)data[i+1]);
				}

				if (size >= i+3) {
					printf(" timeout:%d",(unsigned)data[i+2]);
				}

				if (size >= i+4) {
					printf(" clip:%d",(unsigned)data[i+3]);
				}

				if (size >= i+5) {
					printf(" syncid:%d",(unsigned)data[i+4]);
				}

				if (size >= i+9) {
					printf(" delay:%d",read_be_u32(data+i+5));
				}

				if (size >= i+13) {
					printf(" timeout:%d",read_be_u32(data+i+9));
				}

				if (size >= i+14) {
					printf(" dt:%d",(unsigned)data[i+10]);
				}

				if (size >= i+15) {
					printf(" ...");
				}
			}
			break;
		case MNG_CN_DEFI :
			printf(" id:%d", read_be_u16(data+0));
			if (size >= 3) {
				switch (data[2]) {
					case 0 : printf(" visible:yes"); break;
					case 1 : printf(" visible:no"); break;
					default : printf(" visible:?"); break;
				}
			}
			if (size >= 4) {
				switch (data[3]) {
					case 0 : printf(" concrete:abstract"); break;
					case 1 : printf(" concrete:concrete"); break;
					default : printf(" concrete:?"); break;
				}
			}
			if (size >= 12) {
				printf(" x:%d y:%d", read_be_u32(data + 4), read_be_u32(data + 8) );
			}
			if (size >= 28) {
				printf(" left:%d right:%d top:%d bottom:%d", read_be_u32(data + 12), read_be_u32(data + 16), read_be_u32(data + 20), read_be_u32(data + 24) );
			}
		break;
		case MNG_CN_MOVE :
			printf(" id_from:%d id_to:%d", read_be_u16(data+0), read_be_u16(data+2));
			switch (data[4]) {
				case 0 : printf(" type:replace"); break;
				case 1 : printf(" type:add"); break;
				default : printf(" type:?"); break;
			}
			printf(" x:%d y:%d", read_be_u32(data + 5), read_be_u32(data + 9) );
			break;
		case MNG_CN_PPLT :
			switch (data[0]) {
				case 0 : printf(" type:replacement_rgb"); break;
				case 1 : printf(" type:delta_rgb"); break;
				case 2 : printf(" type:replacement_alpha"); break;
				case 3 : printf(" type:delta_alpha"); break;
				case 4 : printf(" type:replacement_rgba"); break;
				case 5 : printf(" type:delta_rgba"); break;
				default : printf(" type:?"); break;
			}
			i = 1;
			while (i<size) {
				unsigned ssize;
				printf(" %d:%d", (unsigned)data[i], (unsigned)data[i+1]);
				if (data[0] == 0 || data[1] == 1)
					ssize = 3;
				else if (data[0] == 2 || data[1] == 3)
					ssize = 1;
				else
					ssize = 4;
				i += 2 + (data[i+1] - data[i] + 1) * ssize;
			}
			break;
		case PNG_CN_IHDR :
			printf(" width:%d height:%d", read_be_u32(data), read_be_u32(data + 4) );
			printf(" depth:%d", (unsigned)data[8] );
			printf(" color_type:%d", (unsigned)data[9] );
			printf(" compression:%d", (unsigned)data[10] );
			printf(" filter:%d", (unsigned)data[11] );
			printf(" interlace:%d", (unsigned)data[12] );
		break;
	}

	printf("\n");
}

int process(const char* file) {
	FILE* f;
	int len;
	struct stat st;

	if (stat(file,&st) != 0) {
		return -1;
	}

	f = fopen(file,"rb");
	if (!f) {
		return -1;
	}

	len = st.st_size - 8;
	if (fskip(f,8) != 0)
		return -1;
	
	while (len > 0) {
		char v[4];
		unsigned l;
		unsigned c;
		char* data;
		if (fread_be_u32(f,&l) != 0)
			return -1;
		if (fread_tag(f,v) != 0)
			return -1;
		if (l) {
			data = malloc(l);
			if (!data)
				return -1;
			if (fread(data,l,1,f) != 1)
				return -1;
			print(v,data,l);
			free(data);
		} else {
			print(v,0,0);
		}
		if (fread_be_u32(f,&c) != 0)
			return -1;
                len -= 4 + 4 + l + 4;
	}

	if (len != 0)
		return -1;

	fclose(f);

	return 0;
}

int main(int argc, char* argv[]) {
	if (argc!=2) {
		fprintf(stderr,"Syntax: catmng FILE\n");
		exit(EXIT_FAILURE);
	}

	if (process(argv[1]) != 0) {
		fprintf(stderr,"Error reading %s\n",argv[1]);
		exit(EXIT_FAILURE);
	}

	return EXIT_SUCCESS;
}
